## Problem (101. Symmetrical binary tree)

https://leetcode.com/problems/symmetric-tree/

## Title description

```
Given a binary tree, check whether it is mirror symmetrical.



For example, a binary tree [1,2,2,3,4,4,3] is symmetrical.

1
/ \
2 2
/ \ / \
3 4 4 3


But the following [1,2,2,null,3,null,3] is not mirror symmetrical:

1
/ \
2 2
\ \
3 3


Advanced：

Can you use recursion and iteration to solve this problem?


```

## Company

-Ali
-Tencent
-Baidu
-Byte

- bloomberg
- linkedin
- microsoft

## Pre-knowledge

-[Binary tree](https://github.com/azl397985856/leetcode/blob/master/thinkings/basic-data-structure.md) -[recursion](https://github.com/azl397985856/leetcode/blob/master/thinkings/dynamic-programming.md)

## Idea

When I saw this question, my first instinct was DFS. Then I thought: `If the left subtree is a mirror image, and the right subtree is also a mirror image, does it mean that the whole is a mirror image? `. After a few seconds of thinking, this is obviously wrong and does not meet the meaning of the question.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlu96e83wj31200iugme.jpg)

Obviously, the nodes in the left subtree will be compared with the nodes in the right subtree. I have distinguished the colors of the compared elements for your convenience.

My idea here is: `When traversing each node, if I can know who its corresponding symmetrical node is by some method, then I can directly compare whether the two are consistent. `

Therefore, the idea is to traverse twice. During the first traversal, the traversal results are stored in the hash table at the same time, and then the second traversal goes to the hash table to fetch. This method is feasible, but it requires N space (N is the total number of nodes). I thought that if the two can be traversed at the same time, wouldn't the overhead of the hash table be eliminated?

![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlu9a7sy7j31a30u0408.jpg)

If you don't understand, let me give a simple example：

```
Given an array, check if it is mirror symmetrical. For example, the array [1,2,2,3,2,2,1] is symmetrical.
```

If you use a hash table, it is probably：

```py
seen = dict()
for i, num in enumerate(nums):
seen[i] = num
for i, num in enumerate(nums):
if seen[len(nums) - 1 - i] ! = num:
return False
return True
```

And traversing at the same time is probably like this：

```py
l = 0
r = len(nums) - 1

while l < r:
if nums[l] ! = nums[r]: return False
l += 1
r -= 1
return True

```

> In fact, if it is more like this topic, it should be expanded from the middle to both sides.

## Code

Code support: C++, Java, Python3

C++ Code:

```c++
/**
* Definition for a binary tree node.
* struct TreeNode {
* int val;
* TreeNode *left;
* TreeNode *right;
* TreeNode(int x) : val(x), left(NULL), right(NULL) {}
* };
*/
class Solution {
public:
bool isSymmetric(TreeNode* root) {
return root==NULL? true:recur(root->left, root->right);
}

bool recur(TreeNode* l, TreeNode* r)
{
if(l == NULL && r==NULL)
{
return true;
}
// There is only one child node or the left and right are not equal
if(l==NULL || r==NULL || l->val ! = r->val)
{
return false;
}

return recur(l->left, r->right) && recur(l->right, r->left);
}
};
```

Java Code:

```java
/**
* Definition for a binary tree node.
* public class TreeNode {
* int val;
* TreeNode left;
* TreeNode right;
* TreeNode(int x) { val = x; }
* }
*/
class Solution {
public boolean isSymmetric(TreeNode root) {
if(root == null)
{
return true;
}
else{
return recur(root. left, root. right);
}
// return root == null ? true : recur(root. left, root. right);
}

public boolean recur(TreeNode l, TreeNode r)
{
if(l == null && r==null)
{
return true;
}
// There is only one child node or the left and right are not equal
if(l==null || r==null || l. val ! = r. val)
{
return false;
}

return recur(l. left, r. right) && recur(l. right, r. left);
}
}
```

Python3 Code:

```py

class Solution:
def isSymmetric(self, root: TreeNode) -> bool:
def dfs(root1, root2):
if root1 == root2 == None: return True
if not root1 or not root2: return False
if root1. val ! = root2. val: return False
return dfs(root1. left, root2. right) and dfs(root1. right, root2. left)
if not root: return True
return dfs(root. left, root. right)
```

**Complexity analysis**

-Time complexity:$O(N)$, where N is the number of nodes.
-Spatial complexity: The highest depth of recursion is the number of nodes, so the spatial complexity is $O(N)$, where N is the number of nodes.

If you have any comments on this, please leave me a message. I will check the answers one by one when I have time. For more algorithm routines, you can visit my LeetCode problem solving warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.
You can also pay attention to my public account "Force Buckle Plus" to take you to chew off the hard bone of the algorithm.
![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)

![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlu9b4p9ej30x20iwjtf.jpg)
