## Problem (108. Convert an ordered array to a binary search tree)

https://leetcode.com/problems/convert-sorted-array-to-binary-search-tree/

## Title description

```
Convert an ordered array arranged in ascending order into a highly balanced binary search tree.

In this question, a highly balanced binary tree refers to a binary tree. The absolute value of the height difference between the left and right subtrees of each node does not exceed 1.

example:

Given an ordered array: [-10, -3,0,5,9],

One possible answer is: [0,-3, 9,-10, null,5], which can represent the following highly balanced binary search tree：

0
/ \
-3 9
/ /
-10 5

```

## Pre-knowledge

-[Binary search tree](https://github.com/azl397985856/leetcode/blob/master/thinkings/basic-data-structure.md) -[Balanced Binary tree](https://github.com/azl397985856/leetcode/blob/master/thinkings/basic-data-structure.md) -[recursion](https://github.com/azl397985856/leetcode/blob/master/thinkings/dynamic-programming.md)

## Company

-Ali
-Tencent
-Baidu
-Byte

- airbnb

## Idea

Since the input is an ordered array in ascending order of \*\*. Therefore, choose any point and use it as the root node, the left part of the left node, and the right part of the right node. Therefore, it is easy for us to write recursive code.

The title requirement is a binary search tree with a high degree of balance, so we must take the midpoint. It is not difficult to prove: `Since it is the midpoint, the difference between the left and right parts will not be greater than 1, that is to say, the number of nodes of the left and right subtrees formed by it will differ by at most 1, so the absolute value of the height difference between the left and right subtrees will not exceed 1`.

From an image point of view, it's like you lift a rope, and if you lift it from it, you can minimize the difference in the length of the rope on both sides.

![image.png](https://tva1.sinaimg.cn/large/007S8ZIlly1ghltysdgtvj30nj0hv3z2.jpg)

## Key points

-Find the midpoint

## Code

Code support: JS, C++, Java, Python

JS Code:

```js
var sortedArrayToBST = function (nums) {
	// Since the array is sorted, one idea is to divide the array into two halves, one half is the left subtree and the other half is the right subtree
	// Then use the “recursive nature of the tree” to complete the operation recursively.
	if (nums.length === 0) return null;
	const mid = nums.length >> 1;
	const root = new TreeNode(nums[mid]);

	root.left = sortedArrayToBST(nums.slice(0, mid));
	root.right = sortedArrayToBST(nums.slice(mid + 1));
	return root;
};
```

Python Code:

```py
class Solution:
def sortedArrayToBST(self, nums: List[int]) -> TreeNode:
if not nums: return None
mid = (len(nums) - 1) // 2
root = TreeNode(nums[mid])
root. left = self. sortedArrayToBST(nums[:mid])
root. right = self. sortedArrayToBST(nums[mid + 1:])
return root
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity: Each recursion copies the space of N, so the spatial complexity is $O(N^2)$

However, there is actually no need to open up new space：

C++ Code:

```c++
class Solution {
public:
TreeNode* sortedArrayToBST(vector<int>& nums) {
return reBuild(nums, 0, nums. size()-1);
}

TreeNode* reBuild(vector<int>& nums, int left, int right)
{
// Termination condition: the middle-order traversal is empty
if(left > right)
{
return NULL;
}
// Establish the root node of the current subtree
int mid = (left+right)/2;
TreeNode * root = new TreeNode(nums[mid]);

// Recursion of the lower layer of the left subtree
root->left = reBuild(nums, left, mid-1);
// Recursion of the lower layer of the right subtree
root->right = reBuild(nums, mid+1, right);
// Return to the root node
return root;
}
};
```

Java Code:

```java
class Solution {
public TreeNode sortedArrayToBST(int[] nums) {
return dfs(nums, 0, nums. length - 1);
}

private TreeNode dfs(int[] nums, int lo, int hi) {
if (lo > hi) {
return null;
}
int mid = lo + (hi - lo) / 2;
TreeNode root = new TreeNode(nums[mid]);
root. left = dfs(nums, lo, mid - 1);
root. right = dfs(nums, mid + 1, hi);
return root;
}
}

```

Python Code:

```python
class Solution(object):
def sortedArrayToBST(self, nums):
"""
:type nums: List[int]
:rtype: TreeNode
"""
return self. reBuild(nums, 0, len(nums)-1)

def reBuild(self, nums, left, right):
# Termination condition：
if left > right:
return
# Establish the root node of the current subtree
mid = (left + right)//2
root = TreeNode(nums[mid])
# Recursion of the lower layer of the left and right subtrees
root. left = self. reBuild(nums, left, mid-1)
root. right = self. reBuild(nums, mid+1, right)

return root
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity: Since it is a balanced binary tree, the overhead of the implicit call stack is $O(logN)$

For more questions, please visit my LeetCode questions warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

Public account 【[Force Buckle plus] (https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg )】
Zhihu Column 【[Lucifer-Zhihu] (https://www.zhihu.com/people/lu-xiao-13-70 )】

Pay attention, don't get lost!
