## Problem (136. Numbers that appear only once)

https://leetcode.com/problems/single-number/

## Title description

```
Given an array of non-empty integers, each element appears twice except for one element that appears only once. Find the element that has only appeared once.

description：

Your algorithm should have linear time complexity. Can you achieve it without using extra space?

Example 1:

Input: [2,2,1]
Output: 1
Example 2:

Input: [4,1,2,1,2]
Output: 4

```

## Pre-knowledge

-[Bit operation](https://github.com/azl397985856/leetcode/blob/master/thinkings/bit.md)

## Company

-Ali
-Tencent
-Baidu
-Byte

## Idea

According to the title description, since the conditions that the time complexity must be O(n) and the space complexity must be O(1) are added, the sorting method cannot be used, and the map data structure cannot be used.

We can use the nature of binary XOR to complete it, and XOR all numbers to get the only number that appears.

## Key points

1. XOR nature
   The result of XOR of two numbers "a^b" is the number obtained by calculating each binary bit of a and B. The logic of the operation is
   If the number of the same digit is the same, it is 0, and if it is different, it is 1

2. The law of XOR

-Any number that is XOR by itself is `0`

-Any number is different from 0 or `itself`

3. Many people just remember the nature and law of XOR, but lack of understanding of its essence makes it difficult to think of this solution (I didn't expect it myself).

4. bit operation

## Code

-Language support: JS, C, C++, Java, Python

JavaScrip Code：

```js
/**
 * @param {number[]} nums
 * @return {number}
 */
var singleNumber = function (nums) {
	let ret = 0;
	for (let index = 0; index < nums.length; index++) {
		const element = nums[index];
		ret = ret ^ element;
	}
	return ret;
};
```

C Code：

```c
int singleNumber(int* nums, int numsSize){
int res=0;
for(int i=0;i<numsSize;i++)
{
res ^= nums[i];
}

return res;
}
```

C++ Code：

```C++
class Solution {
public:
int singleNumber(vector<int>& nums) {
auto ret = 0;
for (auto i : nums) ret ^= i;
return ret;
}
};

// C++ one-liner
class Solution {
public:
int singleNumber(vector<int>& nums) {
return accumulate(nums. cbegin(), nums. cend(), 0, bit_xor<int>());
}
};
```

Java Code：

```java
class Solution {
public int singleNumber(int[] nums) {
int res = 0;
for(int n:nums)
{
// XOR
res ^= n;
}
return res;
}
}
```

Python Code:

```python
class Solution:
def singleNumber(self, nums: List[int]) -> int:
single_number = 0
for num in nums:
single_number ^= num
return single_number
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity:$O(1)$

## Extension

There is an array of n elements. Except for two numbers that appear only once, the remaining elements appear twice. Let you find out how many of these two numbers appear only once, which requires a time complexity of O(n) and a fixed amount of memory space to be opened up (regardless of n).

It's the same as above, but this time it's not one number, but two numbers. Or according to the above idea, we will perform an XOR operation for all staff，
The result obtained is the XOR result of those two different numbers that only appear once.

We just talked about that there is a "any number and its own XOR is 0" in the law of Xor. Therefore, our idea is whether we can divide these two different numbers into two groups A and B.
Grouping needs to meet two conditions.

1. Two unique numbers are divided into different groups

2. The same numbers are divided into the same groups

In this way, the two numbers can be obtained by XOR of each set of data.

The key point of the question is how do we group?

Due to the nature of XOR, if the same bit is the same, it is 0, and if it is different, it is 1. The result of our XOR of all numbers must not be 0, which means that at least one digit is 1.

Let's take any one, and the basis for grouping will come, that is, the one you take is divided into 1 group by 0, and the one that is 1 is divided into a group.
This will definitely guarantee`2. The same numbers are divided into the same groups`, will different numbers be divided into different groups? Obviously, of course, we can, so we choose 1, which is
Say that'two unique numbers` must be different in that one, so the two unique elements will definitely be divided into different groups.

Done！

For more questions, please visit my LeetCode questions warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

Pay attention to the official account, work hard to restore the problem-solving ideas in clear and straightforward language, and there are a large number of diagrams to teach you how to recognize routines and brush questions efficiently.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
