# Problem (155. Minimum stack)

https://leetcode.com/problems/min-stack/

# Title description

```
Design a stack that supports push, pop, and top operations and can retrieve the smallest element within a constant time.

Push(x)-pushes the element x into the stack.
pop()-delete the element at the top of the stack.
top()--get the top element of the stack.
getMin()--retrieve the smallest element in the stack.


example:

input：
["MinStack","push","push","push","getMin","pop","top","getMin"]
[[],[-2],[0],[-3],[],[],[],[]]

output：
[null,null,null,null,-3,null,0,-2]

explain：
MinStack minStack = new MinStack();
minStack. push(-2);
minStack. push(0);
minStack. push(-3);
minStack. getMin(); --> Return -3.
minStack. pop();
minStack. top(); --> Return 0.
minStack. getMin(); --> Return -2.


prompt：

pop, top, and getMin operations are always called on a non-empty stack.

```

## Pre-knowledge

-[Stack](https://github.com/azl397985856/leetcode/blob/master/thinkings/basic-data-structure.md)

## Company

- amazon
- bloomberg
- google
- snapchat
- uber
- zenefits
  -Ali
  -Tencent
  -Baidu
  -Byte

## Two stacks

### Idea

We use two stacks：

-A stack stores all the elements. Push and pop are normal operations. This normal stack.
-The other one stores the smallest stack. Every time we push, if it is smaller than the top of the smallest stack, we will push into the smallest stack, otherwise we will not operate.
-Every time we pop, we judge whether it is the same as the top element of the smallest stack. If it is the same, then we can pop off the top element of the smallest stack.

### Key points

-Judgment conditions for pushing to minstack. It should be that the stack is empty or x is less than or equal to the top element of the minstack stack

### Code

- Language support: JS, C++, Java, Python

JavaScript Code:

```js
/**
* initialize your data structure here.
*/
var MinStack = function() {
this. stack = []
this. minStack = []
};

/**
* @param {number} x
* @return {void}
*/
MinStack. prototype. push = function(x) {
this. stack. push(x)
if (this. minStack. length == 0 || x <= this. minStack[this. minStack. length - 1]) {
this. minStack. push(x)
}
};

/**
* @return {void}
*/
MinStack. prototype. pop = function() {
const x = this. stack. pop()
if (x ! == void 0 && x === this. minStack[this. minStack. length - 1]) {
this. minStack. pop()
}
};

/**
* @return {number}
*/
MinStack. prototype. top = function() {
return this. stack[this. stack. length - 1]
};

/**
* @return {number}
*/
MinStack. prototype. min = function() {
return this. minStack[this. minStack. length - 1]
};

/**
* Your MinStack object will be instantiated and called as such:
* var obj = new MinStack()
* obj. push(x)
* obj. pop()
* var param_3 = obj. top()
* var param_4 = obj. min()
*/
```

C++ Code:

```c++
class MinStack {
stack<int> data;
stack<int> helper;
public:
/** initialize your data structure here. */
MinStack() {

}

void push(int x) {
data. push(x);
if(helper. empty() || helper. top() >= x)
{
helper. push(x);
}

}

void pop() {
int top = data. top();
data. pop();
if(top == helper. top())
{
helper. pop();
}

}

int top() {
return data. top();
}

int getMin() {
return helper. top();
}
};

/**
* Your MinStack object will be instantiated and called as such:
* MinStack* obj = new MinStack();
* obj->push(x);
* obj->pop();
* int param_3 = obj->top();
* int param_4 = obj->getMin();
*/
```

Java Code:

```java
public class MinStack {

// Data stack
private Stack<Integer> data;
// Auxiliary stack
private Stack<Integer> helper;

/**
* initialize your data structure here.
*/
public MinStack() {
data = new Stack<>();
helper = new Stack<>();
}

public void push(int x) {
// The auxiliary stack is only added when necessary
data. add(x);
if (helper. isEmpty() || helper. peek() >= x) {
helper. add(x);
}
}

public void pop() {
// Key 3: data must be pop()
if (! data. isEmpty()) {
// Note: Declared as int type, automatic unboxing has been completed here, and it has been converted from Integer to int，
// So the following comparison can use the "==" operator
int top = data. pop();
if(top == helper. peek()){
helper. pop();
}
}
}

public int top() {
if(! data. isEmpty()){
return data. peek();
}
}

public int getMin() {
if(! helper. isEmpty()){
return helper. peek();
}
}
}
```

Python3 Code:

```python
class MinStack:

def __init__(self):
"""
initialize your data structure here.
"""
self. stack = []
self. minstack = []

def push(self, x: int) -> None:
self. stack. append(x)
if not self. minstack or x <= self. minstack[-1]:
self. minstack. append(x)

def pop(self) -> None:
tmp = self. stack. pop()
if tmp == self. minstack[-1]:
self. minstack. pop()

def top(self) -> int:
return self. stack[-1]

def min(self) -> int:
return self. minstack[-1]


# Your MinStack object will be instantiated and called as such:
# obj = MinStack()
# obj. push(x)
# obj. pop()
# param_3 = obj. top()
# param_4 = obj. min()
```

**Complexity analysis**
-Time complexity: O(1)
-Spatial complexity: O(1)

## A stack

### Idea

The intuitive way is to update the minimum value every time you modify the stack (push and pop). Then getMin only needs to return the minimum value we calculated，
top can also directly return to the top element of the stack. This approach requires updating the minimum value every time the stack is modified, so the time complexity is O(n).

![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlucity87j30d609ggls.jpg)

Is there a more efficient algorithm? The answer is yes.

Every time we enter the stack, what we save is no longer the real number, but the difference between it and the current minimum value (the minimum value when the current element is not in the stack).
In this way, when we pop and top, we can get the top element of the stack and add the minimum value of the previous one.
In addition, we update min during push and pop, so that it is easier to get min and return min directly.

> Note that the bold “previous" above is not "current minimum value”

After the above analysis, the key to the problem is transformed into “how to obtain the previous minimum value”. The key point to solve this is to use min.

When pop or top：

-If the top element of the stack is less than 0, it means that the top element of the stack is currently the smallest element. It will affect min when it leaves the stack. We need to update min.
The previous smallest value is “min-top element of the stack”, we need to update the previous minimum value to the current minimum value

> Because when the top element of the stack is added to the stack, it is obtained by "top element of the stack = true value-the smallest element of the previous one"，
> And the true value = min, so it can be concluded that "the smallest element on the previous one = the true value-the top element of the stack"

-If the top element of the stack is greater than 0, it means that it has "no effect" on the minimum value, and the previous minimum value is the previous minimum value.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlucqck9mj30ji0k1gn0.jpg)
![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlucsjh58j30ht0b4mxr.jpg)

### Key points

-The minimum stack should not store the real value, but the difference between the real value and min
When-top involves restoring data, please note that it is the minimum value of the previous one.

### Code

- Language support: JS, C++, Java, Python

Javascript Code:

```js
/*
 * @lc app=leetcode id=155 lang=javascript
 *
 * [155] Min Stack
 */
/**
 * initialize your data structure here.
 */
var MinStack = function () {
	this.stack = [];
	this.minV = Number.MAX_VALUE;
};

/**
 * @param {number} x
 * @return {void}
 */
MinStack.prototype.push = function (x) {
	// update 'min'
	const minV = this.minV;
	if (x < this.minV) {
		this.minV = x;
	}
	return this.stack.push(x - minV);
};

/**
 * @return {void}
 */
MinStack.prototype.pop = function () {
	const item = this.stack.pop();
	const minV = this.minV;

	if (item < 0) {
		this.minV = minV - item;
		return minV;
	}
	return item + minV;
};

/**
 * @return {number}
 */
MinStack.prototype.top = function () {
	const item = this.stack[this.stack.length - 1];
	const minV = this.minV;

	if (item < 0) {
		return minV;
	}
	return item + minV;
};

/**
 * @return {number}
 */
MinStack.prototype.min = function () {
	return this.minV;
};

/**
 * Your MinStack object will be instantiated and called as such:
 * var obj = new MinStack()
 * obj. push(x)
 * obj. pop()
 * var param_3 = obj. top()
 * var param_4 = obj. min()
 */
```

C++ Code:

```c++
class MinStack {
stack<long> data;
long min = INT_MAX;
public:
/** initialize your data structure here. */
MinStack() {

}

void push(int x) {
data. push(x - min);
if(x < min)
{
min = x;
}

}

void pop() {
long top = data. top();
data. pop();
// Update minimum value
if(top < 0)
{
min -= top;
}

}

int top() {
long top = data. top();
// The minimum value is min
if (top < 0)
{
return min;
}
else{
return min+top;
}
}

int getMin() {
return min;
}
};

/**
* Your MinStack object will be instantiated and called as such:
* MinStack* obj = new MinStack();
* obj->push(x);
* obj->pop();
* int param_3 = obj->top();
* int param_4 = obj->getMin();
*/
```

Java Code:

```java
class MinStack {
long min;
Stack<Long> stack;

/** initialize your data structure here. */
public MinStack() {
stack = new Stack<>();
}

public void push(int x) {
if (stack. isEmpty()) {
stack. push(0L);
min = x;
}
else {
stack. push(x - min);
if (x < min)
min = x;
}
}

public void pop() {
long p = stack. pop();

if (p < 0) {
// if (p < 0), the popped value is the min
// Recall p is added by this statement: stack. push(x - min);
// So, p = x - old_min
// old_min = x - p
// again, if (p < 0), x is the min so:
// old_min = min - p
min = min - p;
}
}

public int top() {
long p = stack. peek();

if (p < 0) {
return (int) min;
}
else {
// p = x - min
// x = p + min
return (int) (p + min);
}
}

public int getMin() {
return (int) min;
}
}
```

Python Code:

```python
class MinStack:

def __init__(self):
"""
initialize your data structure here.
"""
self. minV = float('inf')
self. stack = []

def push(self, x: int) -> None:
self. stack. append(x - self. minV)
if x < self. minV:
self. minV = x

def pop(self) -> None:
if not self. stack:
return
tmp = self. stack. pop()
if tmp < 0:
self. minV -= tmp

def top(self) -> int:
if not self. stack:
return
tmp = self. stack[-1]
if tmp < 0:
return self. minV
else:
return self. minV + tmp

def min(self) -> int:
return self. minV



# Your MinStack object will be instantiated and called as such:
# obj = MinStack()
# obj. push(x)
# obj. pop()
# param_3 = obj. top()
# param_4 = obj. min()
```

**Complexity analysis**
-Time complexity: O(1)
-Spatial complexity: O(1)

For more questions, please visit my LeetCode questions warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

Pay attention to the official account, work hard to restore the problem-solving ideas in clear and straightforward language, and there are a large number of diagrams to teach you how to recognize routines and brush questions efficiently.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
