## Problem (283. Move zero)

https://leetcode.com/problems/move-zeroes/

## Title description

```
Given an array of nums, write a function to move all 0s to the end of the array while maintaining the relative order of the non-zero elements.

example:

Input: [0,1,0,3,12]
Output: [1,3,12,0,0]
description:

You must operate on the original array, and you cannot copy additional arrays.
Minimize the number of operations.

```

## Pre-knowledge

-[array](https://github.com/azl397985856/leetcode/blob/master/thinkings/basic-data-structure.md)
-Double pointer

## Company

-Ali
-Tencent
-Baidu
-Byte

- bloomberg
- facebook

## Idea

If the topic does not require modify in-place, we can go through it first and save the ones that contain 0 and the ones that do not contain 0 into two arrays, and then splice the two arrays. However, the topic requires modify in-place, that is, there is no need to use additional storage space. The spatial complexity of the method just now is O(n).

So what if modify in-place reduces the spatial complexity to 1?

In fact, you can use **read and write dual pointers** to do it. Specifically, use a slow pointer to represent a write pointer and a fast pointer to represent a read pointer.

Specifically: the reading pointer keeps moving back. If it encounters a value other than 0, the read value is written to the write pointer, which triggers the write pointer to move (in other cases, the write pointer does not move), and the read pointer goes to the end of the algorithm. After this processing, the final position of the write pointer is preceded by all non-zero numbers, and finally, all the positions after the write pointer can be modified to 0.

## Analysis of key points

-Read and write dual pointers

## Code

-Language support: JS, C++, Java, Python

JavaScript Code:

```js
/**
* @param {number[]} nums
* @return {void} Do not return anything, modify nums in-place instead.
*/
var moveZeroes = function (nums) {
let index = 0;
for (let i = 0; i < nums. length; i++) {
const num = nums[i];
if (num ! == 0) {
nums[index++] = num;
}
}

for (let i = index; i < nums. length; i++) {
nums[index++] = 0;
}
};
```

C++ Code：

> The problem-solving idea is consistent with the JavaScript above, and a little code optimization has been done (non-performance optimization, because the time complexity is O(n))：
> Add a cursor to record the position of the next element to be processed, so that you only need to write a loop once.

```C++
class Solution {
public:
void moveZeroes(vector<int>& nums) {
vector<int>::size_type nonZero = 0;
vector<int>::size_type next = 0;
while (next < nums. size()) {
if (nums[next] ! = 0) {
// Using std::swap() will cause a performance loss of 8ms
// swap(nums[next], nums[nonZero]);
auto tmp = nums[next];
nums[next] = nums[nonZero];
nums[nonZero] = tmp;
++nonZero;
}
++next;
}
}
};
```

Java Code:

```java
class Solution {
public void moveZeroes(int[] nums) {
// Double pointer
int i = 0;
for(int j=0; j<nums. length; j++)
{
// Not 0, move forward
if(nums[j] ! = 0)
{
int temp = nums[i];
nums[i] = nums[j];
nums[j] = temp;
i++;
}
}
}
}
```

Python Code:

```python
class Solution:
def moveZeroes(self, nums: List[int]) -> None:
"""
Do not return anything, modify nums in-place instead.
"""
slow = fast = 0
while fast < len(nums):
if nums[fast] ! = 0:
nums[fast], nums[slow] = nums[slow], nums[fast]
slow += 1
fast += 1
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity:$O(1)$

For more questions, please visit my LeetCode questions warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

Pay attention to the official account, work hard to restore the problem-solving ideas in clear and straightforward language, and there are a large number of diagrams to teach you how to recognize routines and brush questions efficiently.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
