## Problem (342. Power of 4)

https://leetcode.com/problems/power-of-four/

## Title description

```
Given an integer (a 32-bit signed integer), write a function to determine whether it is a power of 4.

Example 1:

Input: 16
Output: true
Example 2:

Input: 5
Output: false
Advanced：
Can you complete this question without using loops or recursion?

```

## Pre-knowledge

-Number theory

## Company

-Baidu

- twosigma

## Idea

The intuitive approach is to keep dividing by 4 until it cannot be divisible, and then determine whether it is 1. The code is as follows：

```js
while (num && num % 4 == 0) {
	num /= 4;
}
return num == 1;
```

But this question has a follow up: "Can you do it without loops/recursion”. Therefore, we need to think differently.

Let's take a look at what the power of 4 looks like with a binary representation.

![263.342.power-of-four-1](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlua1uaopj30j009iwf1.jpg)

Found the law: The binary representation of a power of 4 means that the position of 1 is in the odd position (and not in the lowest position), and the other positions are 0.

We can also find that the power of 2 is characterized by the fact that in addition to the lowest position, there is and only one 1 in other positions (1 can be in any position)

We further analyze that if a number is a power of four, then it only needs to satisfy：

1. Is a power of 2, which guarantees that there is and only one 1 in other positions except for the lowest position.
2. This 1 is not in the even position, it must be in the odd position

For the first point, what if a number is guaranteed to be a power of 2? Obviously, you can't stop dividing by 2 to see if the result is equal to 1, so you can loop.
We can use a trick. If a number n is a power of 2, then n & (n-1) must be equal to 0.，
This can be used as a question of thinking, let's think about it.

For the second point, we can take a special number. For this special number, the odd position is 1, and the even position is 0, and then with this special number
`Sum", if it is equal to itself, then there is no doubt that this 1 is no longer in an even position, but must be in an odd position, because if it is in an even position, the result of "sum" is 0. The title requires that n is a 32-bit signed integer, so our special number should be `010101010101010101010101010101` (no need to count, a total of 32 digits).

![263.342.power-of-four-2](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlua2pq5hj30fi0b0q41.jpg)

As shown in the figure above, 64 is summed with this special number, and the result is itself. 8 is the power of 2, but it is not the power of 4. The result of our search is 0.

In order to reflect our own grid, we can use a calculator to find a number with a relatively high grid. Here I chose hexadecimal, and the result is `0x55555555`.

![263.342.power-of-four](https://tva1.sinaimg.cn/large/007S8ZIlly1ghlua3mzibj30b20d70ua.jpg)

See the code area below for the code.

To be honest, this approach is not easy to think of, in fact, there is another way.
If a number is a power of 4, then it only needs to satisfy：

1. Is a multiple of two
2. Minus 1 is a multiple of three

The code is as follows：

```js
return num > 0 && (num & (num - 1)) === 0 && (num - 1) % 3 === 0;
```

## Key points

-Number theory
-2 power characteristics (mathematical properties and binary representation)
-4 power characteristics (mathematical properties and binary representation)

## Code

Language support: JS, Python

JavaScript Code：

```js
/*
* @lc app=leetcode id=342 lang=javascript
*
* [342] Power of Four
*/
/**
* @param {number} num
* @return {boolean}
*/
var isPowerOfFour = function (num) {
// tag: Number theory

if (num === 1) return true;
if (num < 4) return false;

if ((num & (num - 1)) ! == 0) return false;

return (num & 0x55555555) === num;
};
```

Python Code:

```python
class Solution:
def isPowerOfFour(self, num: int) -> bool:
if num == 1:
return True
elif num < 4:
return False
else:
if not num & (num-1) == 0:
return False
else:
return num & 0x55555555 == num

# Another solution: convert a number into a string with a binary representation, and use the relevant operations of the string to judge
def isPowerOfFour(self, num: int) -> bool:
binary_num = bin(num)[2:]
return binary_num. strip('0') == '1' and len(binary_num) % 2 == 1
```

**Complexity analysis**

-Time complexity:$O(1)$
-Spatial complexity:$O(1)$

For more questions, please visit my LeetCode questions warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

Pay attention to the official account, work hard to restore the problem-solving ideas in clear and straightforward language, and there are a large number of diagrams to teach you how to recognize routines and brush questions efficiently.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
