## Problem (437. Path (III)

https://leetcode.com/problems/path-sum-iii/

## Title description

```
Given a binary tree, each node of it stores an integer value.

Find the sum of the paths and the total number of paths equal to the given value.

The path does not need to start at the root node or end at the leaf node, but the path direction must be downward (only from the parent node to the child node).

The binary tree does not exceed 1000 nodes, and the node value range is an integer of [-1000000, 1000000].

example：

root = [10,5,-3,3,2,null,11,3,-2,null,1], sum = 8

10
/ \
5 -3
/ \ \
3 2 11
/ \ \
3 -2 1

Return 3. Paths with sum equal to 8 have:

1. 5 -> 3
2. 5 -> 2 -> 1
3. -3 -> 11

```

## Pre-knowledge

- hashmap

## Company

-Ali
-Tencent
-Baidu
-Byte

## Idea

This question requires us to solve the path from any node to the descendant nodes and return it to the specified value.
Note that here, it does not necessarily start from the root node, nor does it necessarily end at the leaf node.

A simple idea is to solve it directly recursively. The spatial complexity O(n) and time complexity are between O(nlogn) and O(N^2).，
Specific code：

```js
/**
 * Definition for a binary tree node.
 * function TreeNode(val) {
 * this. val = val;
 * this. left = this. right = null;
 * }
 */
// the number of the paths starting from self
function helper(root, sum) {
	if (root === null) return 0;
	const l = helper(root.left, sum - root.val);
	const r = helper(root.right, sum - root.val);

	return l + r + (root.val === sum ? 1 : 0);
}
/**
 * @param {TreeNode} root
 * @param {number} sum
 * @return {number}
 */
var pathSum = function (root, sum) {
	// Spatial complexity O(n) Time complexity is between O(nlogn) and O(N^2)
	// tag: dfs tree
	if (root === null) return 0;
	// the number of the paths starting from self
	const self = helper(root, sum);
	// we don't know the answer, so we just pass it down
	const l = pathSum(root.left, sum);
	// we don't know the answer, so we just pass it down
	const r = pathSum(root.right, sum);

	return self + l + r;
};
```

However, there is also an algorithm with better spatial complexity, which uses hashmaps to avoid double calculations. The time complexity and spatial complexity are both O(n).
This idea is an upgraded version of'subarray-sum-equals-k`. If you can solve that question O(n), this question will not be very difficult.，
Just replaced the array with a binary tree. For specific ideas, you can see [This topic] (. /560.subarray-sum-equals-k.md )

There is a difference here. Let me explain why there is a 'hashmap[acc] = hashmap[acc] - 1;`， The reason is very simple, that is, when we use DFS, when we go back from the bottom, the value of map should also go back. If you are more familiar with the backtracking method， It should be easy to understand. If you are not familiar with it, you can refer to [this topic] (./46.permutations.md ), this question is through'templist. pop()` is done.

In addition, I drew a picture, I believe you will understand after reading it.

When we execute to the bottom：

![437.path-sum-iii](https://tva1.sinaimg.cn/large/007S8ZIlly1ghludenaf3j30l60cyta7.jpg)

Then go back up：

![437.path-sum-iii-2](https://tva1.sinaimg.cn/large/007S8ZIlly1ghludf311tj30ii0bp0ty.jpg)

It is easy to see that our hashmap should not have the record of the first picture, so it needs to be subtracted.

See the code area below for specific implementation.

## Analysis of key points

-Exchange space for time through hashmap
-For this kind of continuous element summation problem, there is a common idea. You can refer to [This topic] (./560.subarray-sum-equals-k.md )

## Code

-Language support: JS

```js
/*
* @lc app=leetcode id=437 lang=javascript
*
* [437] Path Sum III
*/
/**
* Definition for a binary tree node.
* function TreeNode(val) {
* this. val = val;
* this. left = this. right = null;
* }
*/
function helper(root, acc, target, hashmap) {
// see also : https://leetcode.com/problems/subarray-sum-equals-k/

if (root === null) return 0;
let count = 0;
acc += root. val;
if (acc === target) count++;
if (hashmap[acc - target] ! == void 0) {
count += hashmap[acc - target];
}
if (hashmap[acc] === void 0) {
hashmap[acc] = 1;
} else {
hashmap[acc] += 1;
}
const res =
count +
helper(root. left, acc, target, hashmap) +
helper(root. right, acc, target, hashmap);

// Be careful not to forget here
hashmap[acc] = hashmap[acc] - 1;

return res;
}

var pathSum = function (root, sum) {
const hashmap = {};
return helper(root, 0, sum, hashmap);
};
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity:$O(N)$

For more questions, please visit my LeetCode questions warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

Pay attention to the official account, work hard to restore the problem-solving ideas in clear and straightforward language, and there are a large number of diagrams to teach you how to recognize routines and brush questions efficiently.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
