## Problem (53. Maximum subarray sum)

https://leetcode.com/problems/maximum-subarray/

## Title description

```
Given an array of integers nums, find a contiguous subarray with the largest sum (the subarray contains at least one element) and return its largest sum.

example:

input: [-2,1,-3,4,-1,2,1,-5,4]
Output: 6
Explanation: The largest sum of continuous subarrays [4,-1,2,1] is 6.
Advanced:

If you have implemented a solution with a complexity of O(n), try to solve it using a more subtle partition method.

```

## Pre-knowledge

-[Sliding window](https://github.com/azl397985856/leetcode/blob/master/thinkings/slide-window.md) -[Dynamic planning](https://github.com/azl397985856/leetcode/blob/master/thinkings/dynamic-programming.md)

## Company

-Ali
-Baidu
-Byte
-Tencent

- bloomberg
- linkedin
- microsoft

## Idea

This question solves the continuous maximum sub-sequence sum. The following analyzes different problem-solving ideas from the perspective of time complexity.

#### Solution 1-Violent Solution (violence is a miracle, oh yeah! ）

Under normal circumstances, start with the violent solution analysis, and then carry out step-by-step optimization.

**Original violent solution:**(timeout)

To find the sum of the sub-sequence, then we need to know the position of the beginning and end of the sub-sequence, and then calculate the sum of the sequence between the beginning and the end. Use 2 for loops to enumerate the beginning and end positions of all sub-sequences.
Then use a for loop to solve the sequence sum. The time complexity here is too high,`O(n^3)`.

**Complexity analysis**

-Time complexity:$O(N^3)$, where N is the length of the array
-Spatial complexity:$O(1)$

#### Solution 2-Prefix and + Violent solution

**Optimized violent solution:** (shocked, actually AC)

On the basis of the violent solution, we can optimize the violent solution`O(n^2)` with the prefix sum, where space is exchanged for time.
Here you can use the original array to represent`PrefixSum`, saving space.

Finding the sequence sum can be optimized with the prefix sum (`PrefixSum`), given the beginning and end positions of the sub-sequence`(l, r),`
Then the sequence and'subarraysum=PrefixSum[r]-PrefixSum[l-1];` Use a global variable'maxSum` to compare the sum of the sub-sequences solved each time,`maxSum = max(maxSum, subarraySum)'.

**Complexity analysis**

-Time complexity:$O(N^2)$, where N is the length of the array
-Spatial complexity:$O(N)$

> If you change the original array to represent prefixes and arrays, the spatial complexity is reduced to `O(1)`

However, the time complexity is still too high, and can it be more optimized? The answer is yes, the prefix sum can also be optimized to`O(n)`.

####Solution 3-Optimize the prefix and -from [**@lucifer**] (https://github.com/azl397985856 )

We define the function's(i)`, its function is to calculate the value starting from `0(including 0)` and adding to`i(including i)`.

Then's(j)-S(i-1)`is equal to the value from`i`(including i) to`j` (including j).

We further analyze, in fact, we only need to traverse once to calculate all's(i)`, where'i= 0,1,2,. . . . ,n-1. `
Then we subtract the previous's(k)`, where'k= 0,1,2,. . . , i-1`, the minimum value can be. So we need
Use a variable to maintain this minimum value, and also need a variable to maintain the maximum value.

**Complexity analysis**

-Time complexity:$O(N)$, where N is the length of the array
-Spatial complexity:$O(1)$

#### Solution 4-[Partition Method] (https://www.wikiwand.com/zh-hans/%E5%88%86%E6%B2%BB%E6%B3%95 )

We divide the array "nums" into two parts: left ("left") and right ("right") at the middle position ("m"). Then there is，
`left = nums[0]. . . nums[m-1]` and'return = nums[m + 1]. . . nums[n-1]`

There are three situations where the position of the largest sub-sequence sum is as follows：

1. Consider the intermediate element'nums[m]`, which spans the left and right parts. Here, starting from the intermediate element, find the largest suffix to the left and the largest prefix to the right to maintain continuity.
2. Regardless of the intermediate elements, the sum of the largest sub-sequence appears in the left half, and the sum of the largest sub-sequence on the left is solved recursively.
3. Regardless of the intermediate elements, the sum of the largest sub-sequence appears in the right half, and the sum of the largest sub-sequence on the right is solved recursively.

The sum of the largest sub-sequences in the three cases is obtained separately, and the largest value of the three is the sum of the largest sub-sequences.

For example, as shown in the figure below：
![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghltwlo6jpj31400u0acg.jpg)

**Complexity analysis**

-Time complexity:$O(NlogN)$, where N is the length of the array
-Spatial complexity:$O(logN)$

####Solution 5-[Dynamic Planning] (https://www.wikiwand.com/zh-hans/%E5%8A%A8%E6%80%81%E8%A7%84%E5%88%92 )

The difficulty of dynamic programming is to find the state transition equation，

'dp[i]-represents the maximum sub-sequence sum to the current position i`

The state transition equation is：
`dp[i] = max(dp[i - 1] + nums[i], nums[i])`

Initialization:`dp[0] = nums[0]`

From the state transition equation, we only focus on the value of the previous state, so we don't need to open an array to record the sum of all the sub-sequences of positions, only two variables are required.，

`currMaxSum-the cumulative maximum sum to the current position i`

`maxSum-global maximum sub-sequence sum`:

- `currMaxSum = max(currMaxSum + nums[i], nums[i])`
- `maxSum = max(currMaxSum, maxSum)`

As shown in the figure：
![](https://tva1.sinaimg.cn/large/007S8ZIlly1ghltwmokuuj30pj0h20te.jpg)

**Complexity analysis**

-Time complexity:$O(N)$, where N is the length of the array
-Spatial complexity:$O(1)$

## Analysis of key Points

1. Violent solution, enumerate the combinations of the beginning and end positions of all combinatorial sub-sequences, solve the largest sub-sequence sum, and the optimization can be pre-processed to obtain the prefix sum
2. According to the partition method, the array is divided into three parts from the middle position each time, and the maximum sum of the left and right middle (here is the sub-sequence including the intermediate elements) is obtained separately. Recursion is deep for the left and right, and the maximum value of the three is the current maximum sub-sequence sum.
3. Dynamic planning, find the state transition equation, and find the maximum sum of the current position.

## Code (`Java/Python3/Javascript`)

#### Solution 2- Prefix and + violence

_Java code_

```java
class MaximumSubarrayPrefixSum {
public int maxSubArray(int[] nums) {
int len = nums. length;
int maxSum = Integer. MIN_VALUE;
int sum = 0;
for (int i = 0; i < len; i++) {
sum = 0;
for (int j = i; j < len; j++) {
sum += nums[j];
maxSum = Math. max(maxSum, sum);
}
}
return maxSum;
}
}
```

_Python3 code_ `(TLE)`

```python
import sys
class Solution:
def maxSubArray(self, nums: List[int]) -> int:
n = len(nums)
maxSum = -sys. maxsize
sum = 0
for i in range(n):
sum = 0
for j in range(i, n):
sum += nums[j]
maxSum = max(maxSum, sum)

return maxSum
```

_Javascript code_ from [**@lucifer**](https://github.com/azl397985856)

```javascript
function LSS(list) {
	const len = list.length;
	let max = -Number.MAX_VALUE;
	let sum = 0;
	for (let i = 0; i < len; i++) {
		sum = 0;
		for (let j = i; j < len; j++) {
			sum += list[j];
			if (sum > max) {
				max = sum;
			}
		}
	}

	return max;
}
```

#### Solution Three-Optimize the prefix sum

_Java code_

```java
class MaxSumSubarray {
public int maxSubArray3(int[] nums) {
int maxSum = nums[0];
int sum = 0;
int minSum = 0;
for (int num : nums) {
// prefix Sum
sum += num;
// update maxSum
maxSum = Math. max(maxSum, sum - minSum);
// update minSum
minSum = Math. min(minSum, sum);
}
return maxSum;
}
}
```

_Python3 code_

```python
class Solution:
def maxSubArray(self, nums: List[int]) -> int:
n = len(nums)
maxSum = nums[0]
minSum = sum = 0
for i in range(n):
sum += nums[i]
maxSum = max(maxSum, sum - minSum)
minSum = min(minSum, sum)

return maxSum
```

_Javascript code_ from [**@lucifer**](https://github.com/azl397985856)

```javascript
function LSS(list) {
	const len = list.length;
	let max = list[0];
	let min = 0;
	let sum = 0;
	for (let i = 0; i < len; i++) {
		sum += list[i];
		if (sum - min > max) max = sum - min;
		if (sum < min) {
			min = sum;
		}
	}

	return max;
}
```

#### Solution 4-Partition Method

_Java code_

```java
class MaximumSubarrayDivideConquer {
public int maxSubArrayDividConquer(int[] nums) {
if (nums == null || nums. length == 0) return 0;
return helper(nums, 0, nums. length - 1);
}
private int helper(int[] nums, int l, int r) {
if (l > r) return Integer. MIN_VALUE;
int mid = (l + r) >>> 1;
int left = helper(nums, l, mid - 1);
int right = helper(nums, mid + 1, r);
int leftMaxSum = 0;
int sum = 0;
// left surfix maxSum start from index mid - 1 to l
for (int i = mid - 1; i >= l; i--) {
sum += nums[i];
leftMaxSum = Math. max(leftMaxSum, sum);
}
int rightMaxSum = 0;
sum = 0;
// right prefix maxSum start from index mid + 1 to r
for (int i = mid + 1; i <= r; i++) {
sum += nums[i];
rightMaxSum = Math. max(sum, rightMaxSum);
}
// max(left, right, crossSum)
return Math. max(leftMaxSum + rightMaxSum + nums[mid], Math. max(left, right));
}
}
```

_Python3 code_

```python
import sys
class Solution:
def maxSubArray(self, nums: List[int]) -> int:
return self. helper(nums, 0, len(nums) - 1)
def helper(self, nums, l, r):
if l > r:
return -sys. maxsize
mid = (l + r) // 2
left = self. helper(nums, l, mid - 1)
right = self. helper(nums, mid + 1, r)
left_suffix_max_sum = right_prefix_max_sum = 0
sum = 0
for i in reversed(range(l, mid)):
sum += nums[i]
left_suffix_max_sum = max(left_suffix_max_sum, sum)
sum = 0
for i in range(mid + 1, r + 1):
sum += nums[i]
right_prefix_max_sum = max(right_prefix_max_sum, sum)
cross_max_sum = left_suffix_max_sum + right_prefix_max_sum + nums[mid]
return max(cross_max_sum, left, right)
```

_Javascript code_ from [**@lucifer**](https://github.com/azl397985856)

```javascript
function helper(list, m, n) {
	if (m === n) return list[m];
	let sum = 0;
	let lmax = -Number.MAX_VALUE;
	let rmax = -Number.MAX_VALUE;
	const mid = ((n - m) >> 1) + m;
	const l = helper(list, m, mid);
	const r = helper(list, mid + 1, n);
	for (let i = mid; i >= m; i--) {
		sum += list[i];
		if (sum > lmax) lmax = sum;
	}

	sum = 0;

	for (let i = mid + 1; i <= n; i++) {
		sum += list[i];
		if (sum > rmax) rmax = sum;
	}

	return Math.max(l, r, lmax + rmax);
}

function LSS(list) {
	return helper(list, 0, list.length - 1);
}
```

#### Solution 5-Dynamic Planning

_Java code_

```java
class MaximumSubarrayDP {
public int maxSubArray(int[] nums) {
int currMaxSum = nums[0];
int maxSum = nums[0];
for (int i = 1; i < nums. length; i++) {
currMaxSum = Math. max(currMaxSum + nums[i], nums[i]);
maxSum = Math. max(maxSum, currMaxSum);
}
return maxSum;
}
}
```

_Python3 code_

```python
class Solution:
def maxSubArray(self, nums: List[int]) -> int:
n = len(nums)
max_sum_ending_curr_index = max_sum = nums[0]
for i in range(1, n):
max_sum_ending_curr_index = max(max_sum_ending_curr_index + nums[i], nums[i])
max_sum = max(max_sum_ending_curr_index, max_sum)

return max_sum
```

_Javascript code_ from [**@lucifer**](https://github.com/azl397985856)

```javascript
function LSS(list) {
	const len = list.length;
	let max = list[0];
	for (let i = 1; i < len; i++) {
		list[i] = Math.max(0, list[i - 1]) + list[i];
		if (list[i] > max) max = list[i];
	}

	return max;
}
```

## Extension

-If the array is a two-dimensional array, find the sum of the largest subarrays?
-If the product of the largest sub-sequence is required?

## Similar questions

- [Maximum Product Subarray](https://leetcode.com/problems/maximum-product-subarray/)
- [Longest Turbulent Subarray](https://leetcode.com/problems/longest-turbulent-subarray/)

If you have any comments on this, please leave me a message. I will check the answers one by one when I have time. For more algorithm routines, you can visit my LeetCode problem solving warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.
You can also pay attention to my public account "Force Buckle Plus" to take you to chew off the hard bone of the algorithm.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
