## 题目地址（547. 省份数量）

https://leetcode-cn.com/problems/number-of-provinces/

## 题目描述

```

有 N 个城市，其中一些彼此相连，另一些没有相连。如果城市 A 与城市 B 直接相连，且城市 B 与城市 C 直接相连，那么城市 A 与城市 C 间接相连。

省份是一组直接或间接相连的城市，组内不含其他没有相连的城市。

给你一个 N x N 的矩阵 isConnected ，其中 isConnected[i][j] = 1 表示第 i 个城市和第 j 个城市直接相连，而 isConnected[i][j] = 0 表示二者不直接相连。返回矩阵中省份的数量。

示例 1:

输入:
[[1,1,0],
[1,1,0],
[0,0,1]]
输出: 2
说明：已知城市 0 和城市 1 相连，他们在一个省份。
第 2 个城市自己在一个省份。所以返回 2。
示例 2:

输入:
[[1,1,0],
[1,1,1],
[0,1,1]]
输出: 1
说明：已知城市 0 和城市 1 直接相连，城市 1 和城市 2 直接相连，所以城市 0 和城市 2 间接相连，所以他们三个在一个省份，返回 1。
注意：

N 在[1,200]的范围内。
对于所有城市，有 M[i][i] = 1。
如果有 M[i][j] = 1，则有 M[j][i] = 1。

```

## 前置知识

- 并查集

## 公司

- 阿里
- 腾讯
- 百度
- 字节

## 思路

并查集有一个功能是可以轻松计算出连通分量，然而本题的省份的个数，本质上就是连通分量的个数，因此用并查集可以完美解决。

为了简单更加清晰，我将并查集模板代码单尽量独拿出来。

## 代码

`find`, `union`, `connected` 都是典型的模板方法。 懂的同学可能也发现了，我没有做路径压缩，这直接导致 find union connected 的时间复杂度最差的情况退化到 $O(N)$。

当然优化也不难，我们只需要给每一个顶层元素设置一个 size 用来表示连通分量的大小，这样 union 的时候我们将小的拼接到大的上即可。 另外 find 的时候我们甚至可以路径压缩，将树高限定到常数，这样时间复杂度可以降低到 $O(1)$。

```python
class UF:
    parent = {}
    cnt = 0
    def __init__(self, M):
        n = len(M)
        for i in range(n):
            self.parent[i] = i
            self.cnt += 1

    def find(self, x):
        while x != self.parent[x]:
            x = self.parent[x]
        return x
    def union(self, p, q):
        if self.connected(p, q): return
        self.parent[self.find(p)] = self.find(q)
        self.cnt -= 1
    def connected(self, p, q):
        return self.find(p) == self.find(q)

class Solution:
    def findCircleNum(self, M: List[List[int]]) -> int:
        n = len(M)
        uf = UF(M)
        for i in range(n):
            for j in range(i):
                if M[i][j] == 1:
                    uf.union(i, j)
        return uf.cnt

```

**复杂度分析**

- 时间复杂度：平均 $O(logN)$，最坏的情况是 $O(N)$
- 空间复杂度：我们使用了 parent， 因此空间复杂度为 $O(N)$

## 相关专题

- [并查集专题](https://github.com/azl397985856/leetcode/blob/master/thinkings/union-find.md)

大家对此有何看法，欢迎给我留言，我有时间都会一一查看回答。更多算法套路可以访问我的 LeetCode 题解仓库：https://github.com/azl397985856/leetcode 。 目前已经 36K star 啦。
大家也可以关注我的公众号《力扣加加》带你啃下算法这块硬骨头。
![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
