# Problem (66. Plus one)

https://leetcode.com/problems/plus-one

## Title description

```
Given a non-negative integer represented by a non-empty array of integers, add one to the number.

The highest digit is stored at the first position of the array, and only a single digit is stored for each element in the array.

You can assume that except for the integer 0, this integer will not start with zero.

Example 1:

Input: [1,2,3]
Output: [1,2,4]
Explanation: The input array represents the number 123.
Example 2:

Input: [4,3,2,1]
Output: [4,3,2,2]
Explanation: The input array represents the number 4321.
```

> Lucifer tip: Don't add the array directly, convert it to a number, add it, and then turn it back.

## Pre-knowledge

-Traversal of arrays (forward traversal and reverse traversal)

## Idea

In fact, we can think of this question as elementary school students practicing addition, but now it is a fixed “addition by one”, so we just need to consider how to achieve this addition process through traversal.

For addition, we know that we need to perform operations from low to high, so we only need to perform a reverse traversal of the array.

Pseudo code：

```java
for(int i = n - 1; i > - 1; i --) {
Internal logic
}

```

In terms of internal logic, there are actually three situations：

```
1. The number in the single digit is less than 9
17
+ 1
= 18
2. The single digit is equal to 9, and the other digits can be any number from 0 to 9, but the first digit is not equal to 9.
199
+ 1
= 200

109
+ 1
= 110
3. All digits are 9
99
+ 1
= 100

999
+ 1
= 1000
```

The first case is the simplest, we just need to +1 the last bit of the array

In the second case, there is a little more step: we need to move the carry of the bit forward by one bit and calculate whether there are more carry bits.

The third operation is actually the same as the second, but because we know that the length of the array is fixed, we need to expand the length of the array when we encounter situation three. We just need to add one more digit before the result array.

```js
// First of all, we have to start from the last digit of the array and calculate our new sum
sum = arr[arr. length - 1] + 1

// Next we need to determine whether this new sum exceeds 9
sum > 9 ?

// If it is greater than 9, then we will update this bit to 0 and change the carry value to 1
carry = 1
arr[i] = 0

// If it is not greater than 9, update the last digit to sum and return the array directly
arr[arr. length - 1] = sum
return arr

// Then we have to continue to repeat our previous operation to the penultimate position of the array
. . .

// When we are done, if the sum of the first bit of the array is greater than 0, then we must add a 1 to the first bit of the array.
result = new array with size of arr. length + 1
result[0] = 1
result[1] . . . . . . result[result. length - 1] = 0
```

## Code

Code support: Python3, JS, CPP, Go, PHP

Python3 Code：

```py
class Solution:
def plusOne(self, digits: List[int]) -> List[int]:
carry = 1
for i in range(len(digits) - 1, -1, -1):
digits[i], carry = (carry + digits[i]) % 10, (carry + digits[i]) // 10
return [carry] + digits if carry else digits
```

JS Code：

```js
var plusOne = function (digits) {
	var carry = 1; // We treat the initial + 1 as a single-digit carry
	for (var i = digits.length - 1; i > -1; i--) {
		if (carry) {
			var sum = carry + digits[i];
			digits[i] = sum % 10;
			carry = sum > 9 ? 1 : 0; // Each calculation will update the carry that needs to be used in the next step
		}
	}
	if (carry === 1) {
		digits.unshift(1); // If carry stays at 1 at the end, it means that there is a need for an additional length, so we will add a 1 in the first place.
	}
	return digits;
};
```

CPP Code:

```cpp
class Solution {
public:
vector<int> plusOne(vector<int>& A) {
int i = A. size() - 1, carry = 1;
for (; i >= 0 && carry; --i) {
carry += A[i];
A[i] = carry % 10;
carry /= 10;
}
if (carry) A. insert(begin(A), carry);
return A;
}
};
```

Go code:

```go
func plusOne(digits []int) []int {
for i := len(digits) - 1; i >= 0; i-- {
digits[i]++
if digits[i] ! = 10 { // No carry is generated, return directly
return digits
}
Digits[i] = 0// Generate carry, continue to calculate the next digit
}
// All generate carry
digits[0] = 1
digits = append(digits, 0)
return digits
}
```

PHP code:

```php
class Solution {

/**
* @param Integer[] $digits
* @return Integer[]
*/
function plusOne($digits) {
$len = count($digits);
for ($i = $len - 1; $i >= 0; $i--) {
$digits[$i]++;
if ($digits[$i] ! = 10) {// No carry is generated, return directly
return $digits;
}
$ digits[$i] =0; // Generate carry, continue to calculate the next digit
}
// All generate carry
$digits[0] = 1;
$digits[$len] = 0;
return $digits;
}
}
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity:$O(1)$

## Related topics

-[Interview question 02.05. Linked list summation](https://leetcode.com/problems/sum-lists-lcci /)

If you have any comments on this, please leave me a message. I will check the answers one by one when I have time. For more algorithm routines, you can visit my LeetCode problem solving warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

You can also pay attention to my public account "Force Buckle Plus" to take you to chew off the hard bone of the algorithm.
