# Problem (821. The shortest distance of the character)

https://leetcode.com/problems/shortest-distance-to-a-character

## Title description

```
Given a string S and a character C. Returns an array that represents the shortest distance from each character in the string S to the character C in the string S.

Example 1:

Input: S = "loveleetcode", C = "e"
output: [3, 2, 1, 0, 1, 0, 0, 1, 2, 2, 1, 0]
description:

-The length range of the string S is [1, 10000].
-C is a single character, and it is guaranteed to be a character in the string S.
-All letters in S and C are lowercase letters.

```

## Pre-knowledge

-Traversal of arrays (forward traversal and reverse traversal)

## Idea

This question is for us to ask for the closest distance to the target character to the left or right.

I drew a picture for everyone to understand：

![](https://tva1.sinaimg.cn/large/0081Kckwly1gka46lqwlej30rc0f2tae.jpg)

For example, if we want to find the nearest character e of the first character l, the intuitive idea is to search from left to right, stop when we encounter the character e, compare the distances on both sides, and take a smaller one. As shown in the figure above, l is 3 and c is 2.

If this intuitive idea is expressed in code, it looks like this：

Python Code:

```py
class Solution:
def shortestToChar(self, S: str, C: str) -> List[int]:
ans = []

for i in range(len(S)):
# Expand from i to left to right
l = r = i
# Find the first C to the left
while l > -1:
if S[l] == C: break
l -= 1
# Find the first C to the left
while r < len(S):
if S[r] == C: break
r += 1
# If it is not found to the death, then assign an infinitely large number. Since the data range of the topic is [1, 10000], -10000 or 10000 is enough.
if l == -1: l = -10000
if r == len(S): r = 10000
# Just choose the nearest one
ans. append(min(r - i, i - l))
return ans
```

**Complexity analysis**

-Time complexity:$O(N^2)$
-Spatial complexity:$O(1)$

Since the data range of the topic is $10^4$, there is no problem passing all test cases.

But in fact, we can solve it in a linear time. The key points here are similar to the solution above, and they are traversed at both ends. However, it is no longer a blind search, because doing so will have a lot of unnecessary calculations.

We can use the space-for-time method to solve it. Here I use a solution similar to the monotonic stack to solve it. You can also use other methods. Regarding the techniques of monotonic stacks, I will not expand here. Those who are interested can look forward to my later topics.

```py
class Solution:
def shortestToChar(self, S: str, C: str) -> List[int]:
ans = [10000] * len(S)
stack = []
for i in range(len(S)):
while stack and S[i] == C:
ans[stack. pop()] = i - stack[-1]
if S[i] ! = C:stack. append(i)
else: ans[i] = 0
for i in range(len(S) - 1, -1, -1):
while stack and S[i] == C:
ans[stack. pop()] = min(ans[stack[-1]], stack[-1] - i)
if S[i] ! = C:stack. append(i)
else: ans[i] = 0

return ans
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity:$O(N)$

In fact, we don't need a stack to store at all. The reason is very simple, that is, every time we encounter the target character C, we empty all the stacks, so we can identify it with a variable. Refer to the code area later for details.

> If the stack is not emptied when the target character C is encountered, most of the space in this stack cannot be saved, and vice versa.

## Code

Code support: Python3, Java, CPP, Go, PHP

Python3 Code：

```py
class Solution:
def shortestToChar(self, S: str, C: str) -> List[int]:
pre = -10000
ans = []

for i in range(len(S)):
if S[i] == C: pre = i
ans. append(i - pre)
pre = 20000
for i in range(len(S) - 1, -1, -1):
if S[i] == C: pre = i
ans[i] = min(ans[i], pre - i)
return ans
```

Java Code：

```java
class Solution {
public int[] shortestToChar(String S, char C) {
int N = S. length();
int[] ans = new int[N];
int prev = -10000;

for (int i = 0; i < N; ++i) {
if (S. charAt(i) == C) prev = i;
ans[i] = i - prev;
}

prev = 20000;
for (int i = N-1; i >= 0; --i) {
if (S. charAt(i) == C) prev = i;
ans[i] = Math. min(ans[i], prev - i);
}

return ans;
}
}
```

CPP Code:

```cpp
class Solution {
public:
vector<int> shortestToChar(string S, char C) {
vector<int> ans(S. size(), 0);
int prev = -10000;
for(int i = 0; i < S. size(); i ++){
if(S[i] == C) prev = i;
ans[i] = i - prev;
}
prev = 20000;
for(int i = S. size() - 1; i >= 0; i --){
if(S[i] == C) prev = i;
ans[i] = min(ans[i], prev - i);
}
return ans;
}
};
```

Go Code:

```go
func shortestToChar(S string, C byte) []int {
N := len(S)
ans := make([]int, N)

pre:=-N// Maximum distance
for i := 0; i < N; i++ {
if S[i] == C {
pre = i
}
ans[i] = i - pre
}

pre=N*2// Maximum distance
for i := N - 1; i >= 0; i-- {
if S[i] == C {
pre = i
}
ans[i] = min(ans[i], pre-i)
}
return ans
}

func min(a, b int) int {
if a < b {
return a
}
return b
}
```

PHP Code:

```php
class Solution
{

/**
* @param String $S
* @param String $C
* @return Integer[]
*/
function shortestToChar($S, $C)
{
$N = strlen($S);
$ans = [];

$pre = -$N;
for ($i = 0; $i < $N; $i++) {
if ($S[$i] == $C) {
$pre = $i;
}
$ans[$i] = $i - $pre;
}

$pre = $N * 2;
for ($i = $N - 1; $i >= 0; $i--) {
if ($S[$i] == $C) {
$pre = $i;
}
$ans[$i] = min($ans[$i], $pre - $i);
}
return $ans;
}
}
```

**Complexity analysis**

-Time complexity:$O(N)$
-Spatial complexity:$O(1)$

If you have any comments on this, please leave me a message. I will check the answers one by one when I have time. For more algorithm routines, you can visit my LeetCode problem solving warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

You can also pay attention to my public account "Force Buckle Plus" to take you to chew off the hard bone of the algorithm.
