## Problem (88. Merge two ordered arrays)

https://leetcode.com/problems/merge-sorted-array/

## Title description

```
Given two ordered integer arrays nums1 and nums2, merge nums2 into nums1 to make num1 an ordered array.

description:

The number of elements that initialize nums1 and nums2 is m and n, respectively.
You can assume that nums1 has enough space (the size of the space is greater than or equal to m + n) to hold the elements in nums2.
example:

input:
nums1 = [1,2,3,0,0,0], m = 3
nums2 = [2,5,6], n = 3

Output: [1,2,2,3,5,6]
```

## Company

-Ali
-Tencent
-Baidu
-Byte

- loomberg
- facebook
- microsoft

## Pre-knowledge

-Merge and sort

## Idea

The intuitive approach is to 'insert nums2 to the end of num1, and then sort`

Specific code：

```js
// This solution can't even be used for m
// This is obviously not what the questioner meant
if (n === 0) return;
let current2 = 0;
for (let i = nums1.length - 1; i >= nums1.length - n; i--) {
	nums1[i] = nums2[current2++];
}
nums1.sort((a, b) => a - b); // Of course you can write the sort by yourself, I don't bother to write it here, because I have deviated from the topic itself.
```

This question is actually very similar to `merge sort` in the basic sorting algorithm.

Let's review the merge process of merge sort first. The process of merge `yes' is to first compare the header elements of the two arrays, then push the smaller one into the final array, and queue it out of the original array. Keep looping until both arrays are empty.

The specific code is as follows：

```js
// Merge nums1 and nums2
function merge(nums1, nums2) {
	let ret = [];
	let i = (j = 0);
	while (i < nums1.length || j < nums2.length) {
		if (i === nums1.length) {
			ret.push(nums2[j]);
			j++;
			continue;
		}

		if (j === nums2.length) {
			ret.push(nums1[i]);
			i++;
			continue;
		}
		const a = nums1[i];
		const b = nums2[j];
		if (a > b) {
			ret.push(nums2[j]);
			j++;
		} else {
			ret.push(nums1[i]);
			i++;
		}
	}
	return ret;
}
```

But merge sort Many times, when we merge, we usually create a new array, but this question requires `modify in place'. This is a bit different from the merge process of merge sort. It is required to modify it in place here. If you use a method similar to the above, if you use traversal from scratch, you need to put the first m arrays of nums1 into another array to avoid interference from writing pointers. In this way, the spatial complexity is $O(m)$. In fact, we can just compare from the back to the front and insert it from the back to the front. \*\*

We need three pointers：

1. Write the pointer current, which is used to record that the current position has been filled to that position

2. m is used to record which element has been processed in the nums1 array

3. n is used to record which element has been processed in the nums2 array

As shown in the figure：

-Gray represents the processed elements of the num2 array
-Red represents the element currently being compared
-Green represents the element that is already in place

![88.merge-sorted-array-1](https://tva1.sinaimg.cn/large/007S8ZIlly1ghludbcompj30h00n10tj.jpg)
![88.merge-sorted-array-2](https://tva1.sinaimg.cn/large/007S8ZIlly1ghludbuxg8j30dv08l0sv.jpg)
![88.merge-sorted-array-3](https://tva1.sinaimg.cn/large/007S8ZIlly1ghludcsa7oj30ca077wek.jpg)

## Analysis of key points

-Compare from back to front and insert from back to front, so as to avoid the impact of writing pointers, while reducing the space complexity to $O(1)$

## Code

Code support: Python3, C++, Java, JavaScript

JavaSCript Code:

```js
var merge = function (nums1, m, nums2, n) {
	// Set a pointer, the pointer initialization points to the end of nums1 (according to #62, it should be the position where the index is m+n-1, because the length of nums1 may be longer)
	// Then keep moving the pointer to the left to update the element
	let current = m + n - 1;

	while (current >= 0) {
		// No need to continue
		if (n === 0) return;

		// In order to facilitate everyone's understanding, the code here is a bit redundant
		if (m < 1) {
			nums1[current--] = nums2[--n];
			continue;
		}

		if (n < 1) {
			nums1[current--] = nums1[--m];
			continue;
		}
		// Take the big one to fill the end of nums1
		// Then update m or n
		if (nums1[m - 1] > nums2[n - 1]) {
			nums1[current--] = nums1[--m];
		} else {
			nums1[current--] = nums2[--n];
		}
	}
};
```

C++ code:

```
class Solution {
public:
void merge(vector<int>& nums1, int m, vector<int>& nums2, int n) {
int current = m + n - 1;
while (current >= 0) {
if (n == 0) return;
if (m < 1) {
nums1[current--] = nums2[--n];
continue;
}
if (n < 1) {
nums1[current--] = nums1[--m];
continue;
}
if (nums1[m - 1] > nums2[n - 1]) nums1[current--] = nums1[--m];
else nums1[current--] = nums2[--n];
}
}
};
```

Java Code:

```java
class Solution {
public void merge(int[] nums1, int m, int[] nums2, int n) {
int i=m-1, j=n-1, k=m+n-1;
// Merge
while(i>=0 && j>=0)
{
if(nums1[i] > nums2[j])
{
nums1[k--] = nums1[i--];
}
else
{
nums1[k--] = nums2[j--];
}
}
// Merge the remaining nums2
while(j>=0)
{
nums1[k--] = nums2[j--];
}
}
}
```

Python Code:

```python
class Solution:
def merge(self, nums1: List[int], m: int, nums2: List[int], n: int) -> None:
"""
Do not return anything, modify nums1 in-place instead.
"""
pos = m + n - 1
while m > 0 and n > 0:
if nums1[m - 1] < nums2[n - 1]:
nums1[pos] = nums2[n - 1]
n -= 1
else:
nums1[pos] = nums1[m - 1]
m -= 1
pos -= 1
while n > 0:
nums1[pos] = nums2[n - 1]
n -= 1
pos -= 1

```

**Complexity analysis**

-Time complexity:$O(M +N)$
-Spatial complexity:$O(1)$

If you have any comments on this, please leave me a message. I will check the answers one by one when I have time. For more algorithm routines, you can visit my LeetCode problem solving warehouse:https://github.com/azl397985856/leetcode . There are already 37K stars.

You can also pay attention to my public account "Force Buckle Plus" to take you to chew off the hard bone of the algorithm.

![](https://tva1.sinaimg.cn/large/007S8ZIlly1gfcuzagjalj30p00dwabs.jpg)
